!! Copyright (C) 2004-2010 M. Oliveira, F. Nogueira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!! 02111-1307, USA.

#include "global.h"

module atom
  use global
  use oct_parser
  use io
  use messages
  use units
  use splines
  use output
  use mesh
  use ps_io
  use quantum_numbers
  use potentials
  use wave_equations_integrator
  use hamann
  use eigensolver_m
  use states
  use hartree
  use xc
  use scf
  use ps_generation
  implicit none


                    !---Derived Data Types---!

  type atom_type
    private
    integer :: type
    real(R8) :: z
    character(3) :: symbol
    integer :: wave_eq
    type(xc_type):: xc_model
    type(potential_type) :: potential
    type(mesh_type) :: m
    integer :: nspin
    integer :: n_states, n_sc
    type(state_type), pointer :: states(:,:)
  end type atom_type


                    !---Global Variables---!

  integer, parameter :: M_TEST_LD = 1, &
                        M_TEST_DM = 2

  integer, parameter :: M_AE_ATOM = 1, &
                        M_PS_ATOM = 2, &
                        M_KB_ATOM = 3


                    !---Public/Private Statements---!

  private
  public :: atom_type, &
            atom_null, &
            atom_create_ae, &
            atom_create_ps, &
            atom_save, &
            atom_load, &
            atom_end, &
            atom_solve, &      
            atom_test, &
            atom_output, &
            M_TEST_LD, &
            M_TEST_DM

contains

  subroutine atom_null(atm)
    !-----------------------------------------------------------------------!
    ! Nullifies and sets to zero all the components of the atom atm.        !
    !-----------------------------------------------------------------------!
    type(atom_type), intent(out) :: atm

    atm%type = 0
    atm%z = M_ZERO
    atm%symbol = ""
    atm%wave_eq = 0
    atm%nspin = 0
    atm%n_states = 0
    atm%n_sc = 0
    call xc_null(atm%xc_model)
    call potential_null(atm%potential)
    call mesh_null(atm%m)
    nullify(atm%states)

  end subroutine atom_null

  subroutine atom_create_ae(atm, integrator_sp, integrator_dp, eigensolver)
    !-----------------------------------------------------------------------!
    ! Initializes the atom by reading the relevant input options and prints !
    ! some relevant informantion.                                           !
    !-----------------------------------------------------------------------!
    type(atom_type),        intent(inout) :: atm
    type(integrator_type),  intent(inout) :: integrator_sp, integrator_dp
    type(eigensolver_type), intent(in)    :: eigensolver

    integer :: order
    real(R8) :: etot, ekin, eh, enuc, exc
    real(R8), allocatable :: vh(:,:), vxc(:,:), vxctau(:,:)

    call push_sub("atom_create_ae")

                    !---Read information about the atom---!

    !Read nuclear charge
    call oct_parse_double('NuclearCharge', M_ZERO, atm%z)
    if (atm%z <= M_ZERO) then
      message(1) = "Wrong or absent value for the nuclear charge."
      call write_fatal(1)
    end if
    atm%symbol = label(atm%z)

    !Read rel
    call  oct_parse_int('WaveEquation', SCHRODINGER, atm%wave_eq)
    select case (atm%wave_eq)
    case (SCHRODINGER, DIRAC, SCALAR_REL)
    case default
      message(1) = "Illegal WaveEquation."
      call write_fatal(1)
    end select

    !Read polarization
    call oct_parse_int('SpinMode', 1, atm%nspin)
    select case (atm%nspin)
    case (1)
    case (2)
      if (atm%wave_eq == DIRAC) then
        message(1) = "Relativistic SDFT not yet implemented."
        call write_fatal(1)
      end if
    case default
      message(1) = "Illegal SpinMode."
      call write_fatal(1)
    end select


                    !---Initialize the data structures---!

    atm%type = M_AE_ATOM

    !Initialize xc
    call xc_null(atm%xc_model)
    call xc_init(atm%nspin, atm%xc_model)

    !Initialize mesh
    call mesh_null(atm%m)
    call mesh_init(atm%z, atm%m)

    !Initialize orbitals from file
    call atom_init_states_from_block(atm, "Orbitals")

    !Initialize potential
    allocate(vh(atm%m%np, atm%nspin), vxc(atm%m%np, atm%nspin), vxctau(atm%m%np, atm%nspin))
    call hartree_potential(atm%m, atm%nspin, state_density(atm%nspin, atm%n_states, atm%states, atm%m), vh = vh)
    call xc_potential(atm%xc_model, atm%m, atm%nspin, atm%n_states, atm%states, &
                      vxc = vxc, vxctau=vxctau)
    call potential_init(atm%potential, atm%m, atm%nspin, vh + vxc, vxctau, atm%z)
    deallocate(vh, vxc, vxctau)


                    !---Print information about the atom---!

    message(1) = ""
    message(2) = "General Information about the atom:"
    call write_info(2)
    call write_info(2,unit=out_ae)

    !Print info to screen and to the ae/info file
    if     (atm%wave_eq == SCHRODINGER .and. atm%nspin == 1) then
      write(message(1),'(2X,"Non relativistic spin-unpolarized calculations for ",A)') trim(atm%symbol)
    elseif (atm%wave_eq == DIRAC       .and. atm%nspin == 1) then
      write(message(1),'(2X,"Relativistic spin-unpolarized calculations for ",A)') trim(atm%symbol)
    elseif (atm%wave_eq == SCALAR_REL  .and. atm%nspin == 1) then
      write(message(1),'(2X,"Scalar-relativistic spin-unpolarized calculations for ",A)') trim(atm%symbol)
    elseif (atm%wave_eq == SCHRODINGER .and. atm%nspin == 2) then
      write(message(1),'(2X,"Non relativistic spin-polarized calculations for ",A)') trim(atm%symbol)
    elseif (atm%wave_eq == DIRAC       .and. atm%nspin == 2) then
      write(message(1),'(2X,"Relativistic spin-polarized calculations for ",A)') trim(atm%symbol)
    elseif (atm%wave_eq == SCALAR_REL  .and. atm%nspin == 2) then
      write(message(1),'(2X,"Scalar-relativistic spin-polarized calculations for ",A)') trim(atm%symbol)
    end if
    write(message(2),'(2X,"Nuclear charge   : ",F6.2)') atm%z
    write(message(3),'(2X,"Total charge     : ",F6.2)') atm%z - sum(state_charge(atm%states))  

    call write_info(3, verbose_limit=20)
    call write_info(3, unit=out_ae)

    call states_output_configuration(atm%states, atm%nspin, atm%n_states, verbose_limit=20)
    call states_output_configuration(atm%states, atm%nspin, atm%n_states, unit=out_ae)

    call xc_model_output_info(atm%xc_model, verbose_limit=20)
    call xc_model_output_info(atm%xc_model, unit=out_ae)

    call mesh_output_params(atm%m, verbose_limit=20)
    call mesh_output_params(atm%m, unit=out_ae)


                    !---Solve Khon-Sham equation---!

    call atom_solve(atm, integrator_sp, integrator_dp, eigensolver, out_ae)

    !Output info to info file
    write(out_ae,*)
    write(out_ae,'("Density Moments")')
    do order = -2, 5
      if (order == 0) cycle
      write(out_ae,'("  <r^",I2,"> = ",F12.3)') order, & 
           state_density_moment(atm%nspin, atm%n_states, atm%states, atm%m, order)/&
           units_out%length%factor**order
    end do
    write(out_ae,*)
    call atom_energies(atm, etot, ekin, eh, enuc, exc)
    write(out_ae,'("Energies [",A,"]")') trim(units_out%energy%abbrev)
    write(out_ae,'("  Total energy:                         ",F14.6)') etot/units_out%energy%factor
    write(out_ae,'("  Kinetic energy:                       ",F14.6)') ekin/units_out%energy%factor
    write(out_ae,'("  Electron-electron interaction energy: ",F14.6)') eh/units_out%energy%factor
    write(out_ae,'("  Electron-nucleus interaction energy:  ",F14.6)') enuc/units_out%energy%factor
    write(out_ae,'("  Exchange-correlation energy:          ",F14.6)') exc/units_out%energy%factor
    write(out_ae,*)
    call states_output_eigenvalues(atm%nspin, atm%n_states, atm%states, .true., out_ae)

    call pop_sub()
  end subroutine atom_create_ae

  subroutine atom_init_states_from_block(atm, block_name)
    !-----------------------------------------------------------------------!
    ! Initializes the atom by reading the relevant input options and prints !
    ! some relevant informantion.                                           !
    !-----------------------------------------------------------------------!
    type(atom_type),  intent(inout) :: atm
    character(len=*), intent(in)    :: block_name

    type input_data
      integer  :: n
      integer  :: l
      real(R8) :: occ
    end type input_data

    integer :: n_lines, n_id, n_rare, l1, i, p, k, m, n, n_cols, ierr, j_dim
    integer(POINTER_SIZE) :: blk
    real(R8) :: occ, s, j
    character(2) :: rare
    character(5) :: label
    type(qn_type) :: qn
    type(input_data), allocatable :: id(:,:)
    type(input_data), parameter :: ID_RARE(15) = &
         (/ input_data(1, 0, M_TWO), input_data(2, 0, M_TWO), &
            input_data(2, 1, M_SIX), input_data(3, 0, M_TWO), &
            input_data(3, 1, M_SIX), input_data(3, 2, M_TEN), &
            input_data(4, 0, M_TWO), input_data(4, 1, M_SIX), &
            input_data(4, 2, M_TEN), input_data(5, 0, M_TWO), &
            input_data(5, 1, M_SIX), input_data(4, 3, M_FOURTEEN), &
            input_data(5, 2, M_TEN), input_data(6, 0, M_TWO), &
            input_data(6, 1, M_SIX) /)    

    call push_sub("atom_init_states_from_block")

    !Open the orbitals block and get the number of lines
    ierr = oct_parse_block(trim(block_name), blk)
    n_lines = oct_parse_block_n(blk)
    if (n_lines < 1) then
      write(message(1),'("No ",A," block specified on input.")') trim(block_name)
      call write_fatal(1)
    end if

    !Do we have a rare gas in the orbitals block?
    if (oct_parse_block_cols(blk, 0) == 1) then
      call oct_parse_block_string(blk, 0, 0, rare)
      select case (rare)
      case ("He") 
        n_rare = 1
      case ("Ne") 
        n_rare = 3
      case ("Ar") 
        n_rare = 5
      case ("Kr") 
        n_rare = 8
      case ("Xe") 
        n_rare = 11
      case ("Rn") 
        n_rare = 15
      case default
        message(1) = "Wrong input for orbitals block: unknown rare gas."
        call write_fatal(1)
      end select
      l1 = 1
      n_id = n_rare + n_lines - 1
    else
      n_rare = 0
      l1 = 0
      n_id = n_lines
    end if

    allocate(id(atm%nspin, n_id))
    do p = 1, atm%nspin
      id(p, 1:n_rare) = ID_RARE(1:n_rare)
      id(p, 1:n_rare)%occ = id(p, 1:n_rare)%occ/real(atm%nspin, R8)
    end do

    k = n_rare
    do i = l1, n_lines - 1
      k = k + 1

      !Read input
      n_cols = oct_parse_block_cols(blk, i)
      call oct_parse_block_int(blk, i, 0, id(1, k)%n)
      call oct_parse_block_int(blk, i, 1, id(1, k)%l)
      call oct_parse_block_double(blk, i, 2, id(1, k)%occ)
      select case (n_cols)
      case (3)
        if (atm%nspin == 2) then
          id(1, k)%occ = id(1, k)%occ/M_TWO
          id(2, k) = id(1, k)
        end if
      case (4)
        if (atm%nspin == 1) then
          call oct_parse_block_double(blk, i, 3, occ)
          id(1, k)%occ = id(1, k)%occ + occ
        else
          id(2, k) = id(1, k)
          call oct_parse_block_double(blk, i, 3, id(2, k)%occ)
        end if
      case default
        write(message(1),'("There is something wrong at line ",I2," of Orbitals block.")') i+1
        message(2) = "Illegal number of columns."
        call write_fatal(2)        
      end select

      !Check input
      if (id(1, k)%n < 1) then
        write(message(1),'("There is something wrong at line ",I2," of Orbitals block.")') i+1
        message(2) = "Illegal main quantum number."
        call write_fatal(2)
      else if (id(1, k)%l < 0 .or. id(1, k)%l > id(1, k)%n - 1) then
        write(message(1),'("There is something wrong at line ",I2," of Orbitals block.")') i+1
        message(2) = "Illegal angular momentum quantum number."
        call write_fatal(2)
      else if ( (sum(id(:, k)%occ) > id(1, k)%l*M_FOUR + M_TWO).or. &
              (any(id(:, k)%occ > (id(1, k)%l*M_FOUR + M_TWO)/atm%nspin)) ) then
        write(message(1),'("There is something wrong at line ",I2," of Orbitals block")') i+1
        message(2) = "Illegal occupancies."
        call write_fatal(2)
      end if

    end do

    !Close the Orbitals block
    call oct_parse_block_end(blk)

    !Check that there are no duplicated lines
    do i = 1, n_id - 1
      do k = i + 1, n_id
        if (id(1, i)%n == id(1, k)%n .and. id(1, i)%l == id(1, k)%l) then
          message(1) = "There are at least two lines with the same set of quantum numbers."
          message(2) = "in the Orbitals block."
          call write_fatal(2)
        end if
      end do
    end do

    !Initialize states
    atm%n_states = 0
    do i = 1, n_id
      if (atm%wave_eq == DIRAC .and. id(1, i)%l /= M_ZERO) then
        atm%n_states = atm%n_states + 2
      else
        atm%n_states = atm%n_states + 1
      end if
    end do
    allocate(atm%states(atm%nspin, atm%n_states))
    do p = 1, atm%nspin
      do i = 1, atm%n_states
        call state_null(atm%states(p, i))
      end do
    end do

    do p = 1, atm%nspin
      if (atm%nspin == 1) then
        s = M_ZERO
      else
        s = p - M_THREE/M_TWO
      end if

      k = 0
      do i = 1, n_id
        j_dim = 1
        if (atm%wave_eq == DIRAC .and. id(p, i)%l /= 0) j_dim = 2
        do m = 1, j_dim          
          k = k + 1

          if (atm%wave_eq == DIRAC) then
            if (id(p, i)%l == 0) then
              j = M_HALF
              occ = id(p, i)%occ
            else
              j = id(p, i)%l + m - M_THREE/M_TWO
              occ = id(p, i)%occ*(M_TWO*j + M_ONE)/(M_FOUR*id(p, i)%l + M_TWO)
            end if
          else
            j = M_ZERO
            occ = id(p, i)%occ
          end if
          label = qn_label(qn_init(id(p, i)%n, id(p, i)%l, s, j), spin=.false.)
          if (atm%type == M_PS_ATOM) then
            n = id(p, i)%n - &
               minval(id(p, :)%n, mask=id(p, :)%l == id(p, i)%l) + id(p, i)%l + 1
          else
            n = id(p, i)%n
          end if
          qn = qn_init(n, id(p, i)%l, s, j)

          call state_init(atm%states(p, k), atm%m, qn, occ, label)
        end do
      end do

    end do
    deallocate(id)

    call pop_sub()
  end subroutine atom_init_states_from_block

  subroutine atom_create_ps(ae_atm, ps_atm, kb_atm, integrator_sp, integrator_dp, eigensolver)
    !-----------------------------------------------------------------------!
    ! Creates a pseudo-atom and the corresponding KB atom.                  !
    !-----------------------------------------------------------------------!
    type(atom_type),        intent(in)    :: ae_atm
    type(atom_type),        intent(inout) :: ps_atm, kb_atm
    type(integrator_type),  intent(inout) :: integrator_sp, integrator_dp
    type(eigensolver_type), intent(in)    :: eigensolver

    integer :: i, p, file_format
    real(R8) :: z_val
    type(ps_generator_type) :: ps_gen

    call push_sub("atom_create_ps")

    ASSERT(ae_atm%type == M_AE_ATOM)

    ps_atm%type = M_PS_ATOM
    kb_atm%type = M_KB_ATOM

    !Get information from all-electron atom
    !(except the atomic number, which will be obtained later)
    ps_atm%symbol = ae_atm%symbol
    ps_atm%wave_eq = ae_atm%wave_eq
    ps_atm%nspin = ae_atm%nspin
    ps_atm%xc_model = ae_atm%xc_model
    ps_atm%m = ae_atm%m

    kb_atm%symbol = ae_atm%symbol
    kb_atm%wave_eq = ae_atm%wave_eq
    kb_atm%nspin = ae_atm%nspin
    kb_atm%xc_model = ae_atm%xc_model
    kb_atm%m = ae_atm%m

    !Initialize pseudopotentials generator
    call ps_generator_null(ps_gen)
    call ps_generator_init(ps_atm%nspin, ps_atm%wave_eq, ps_gen, &
                           ps_atm%n_states, ps_atm%n_sc)

    !Allocate memory
    allocate(ps_atm%states(ps_atm%nspin, ps_atm%n_states))
    do p = 1, ps_atm%nspin
      do i = 1, ps_atm%n_states
        call state_null(ps_atm%states(p, i))
      end do
    end do

    !Generate pseudopotentials
    call ps_generate(ps_gen, ps_atm%m, ps_atm%wave_eq, ps_atm%nspin, &
         ae_atm%n_states, ae_atm%states, ae_atm%potential, ps_atm%xc_model, &
         integrator_sp, integrator_dp, eigensolver, ps_atm%states, &
         ps_atm%potential)

    !Get the correct atomic number
    z_val = sum(state_charge(ps_atm%states))
    ps_atm%z = ae_atm%z - sum(state_charge(ae_atm%states)) + z_val
    kb_atm%z = ps_atm%z

    !Generate KB projectors
    call ps_generate_kb(ps_gen, ps_atm%m, ps_atm%wave_eq, ps_atm%nspin, &
         ps_atm%n_states, ps_atm%n_sc, ps_atm%states, ps_atm%z, &
         ps_atm%xc_model, ps_atm%potential, integrator_sp, integrator_dp, &
         eigensolver, kb_atm%potential)

    !Save pseudopotentials to output file
    call oct_parse_int('PPOutputFileFormat', M_UPF, file_format)
    call ps_io_init(ps_atm%m, ps_atm%wave_eq, ps_atm%nspin, ps_gen%scheme, &
                    ae_atm%z, ps_atm%z, z_val, ps_atm%symbol, file_format)
    call potential_ps_io_set(ps_atm%potential)
    call potential_ps_io_set(kb_atm%potential)
    call states_ps_io_set(ps_atm%m, ps_atm%nspin, ps_atm%n_states, &
         ps_atm%states, ps_gen%rc)
    call xc_ps_io_set(ps_atm%xc_model)
    call ps_io_save()

    call ps_generator_end(ps_gen)

    call pop_sub()
  end subroutine atom_create_ps

  subroutine atom_save(atm, dir)
    !-----------------------------------------------------------------------!
    ! Writes the atom information to the "dir/data" file.                   !
    !-----------------------------------------------------------------------!
    type(atom_type),  intent(in) :: atm
    character(len=*), intent(in) :: dir

    integer :: unit, iostat, p, i
    character(len=20) :: filename

    call push_sub("atom_save")

    call io_assign(unit)
    filename = trim(dir)//"/data"
    open(unit, file=trim(filename), form='unformatted', status='unknown', iostat=iostat)
    if (iostat > 0) then
      write(message(1),'("Unable to open file ''A''")') trim(filename)
      call write_fatal(1)
    end if

    write(unit) atm%type, atm%z, atm%symbol, atm%wave_eq, atm%nspin, &
                atm%n_states, atm%n_sc

    !Write exchange-correlation model
    call xc_model_save(unit, atm%xc_model)

    !Write mesh
    call mesh_save(unit, atm%m)

    !Write states info
    do p = 1, atm%nspin
      do i = 1, atm%n_states
        call state_save(unit, atm%states(p, i)) 
      end do
    end do

    !Write potential
    call potential_save(unit, atm%potential)

    close(unit)

    call pop_sub()
  end subroutine atom_save

  subroutine atom_load(atm, dir)
    !-----------------------------------------------------------------------!
    ! Reads the atom information from the "dir/data" file.                  !
    !-----------------------------------------------------------------------!
    type(atom_type), intent(inout) :: atm
    character(len=*), intent(in) :: dir

    integer :: unit, iostat, p, i
    character(len=20) :: filename

    call push_sub("atom_load")

    !Open file
    call io_assign(unit)
    filename = trim(dir)//"/data"
    open(unit, file=trim(filename), form='unformatted', status='old', iostat=iostat)
    if (iostat > 0) then
      write(message(1),'("Unable to open file ''A''")') trim(filename)
      call write_fatal(1)
    end if

    !
    read(unit) atm%type, atm%z, atm%symbol, atm%wave_eq, atm%nspin, &
               atm%n_states, atm%n_sc

    !Read exchange-correlation model
    call xc_model_load(unit, atm%xc_model)

    !Read mesh
    call mesh_load(unit, atm%m)

    !Read states data
    allocate(atm%states(atm%nspin, atm%n_states))
    do p = 1, atm%nspin
      do i = 1, atm%n_states
        call state_load(unit, atm%states(p, i))
      end do
    end do

    !Read potential
    call potential_load(unit, atm%potential)

    close(unit)

    call pop_sub()
  end subroutine atom_load

  subroutine atom_end(atm)
    !-----------------------------------------------------------------------!
    ! Frees all the memory associated to atm.                               !
    !-----------------------------------------------------------------------!
    type(atom_type), intent(inout) :: atm

    integer :: p, i

    call push_sub("atom_end")

    atm%type = 0
    atm%z = M_ZERO
    atm%symbol = ""
    atm%wave_eq = 0
    atm%nspin = 0
    atm%n_states = 0
    atm%n_sc = 0

    !End potential
    call potential_end(atm%potential)

    !End xc
    call xc_end(atm%xc_model)

    !End mesh
    call mesh_end(atm%m)

    !End states
    if (associated(atm%states)) then
      do p = 1, atm%nspin      
        do i = 1, atm%n_states
          call state_end(atm%states(p,i))
        end do
      end do
      deallocate(atm%states)
    end if

    call pop_sub()
  end subroutine atom_end

  subroutine atom_solve(atm, integrator_sp, integrator_dp, eigensolver, unit)
    !-----------------------------------------------------------------------!
    ! Solve the Kohn-Sham equations for the atom in a self-consitent way an !
    ! outputs the relevant results.                                         !
    !                                                                       !
    !  atm           - atom object                                          !
    !  integrator_sp - single-precision integrator object                   !
    !  integrator_dp - double-precision integrator object                   !
    !  unit          - output file                                          !
    !-----------------------------------------------------------------------!
    type(atom_type),        intent(inout) :: atm
    type(integrator_type),  intent(inout) :: integrator_sp, integrator_dp
    type(eigensolver_type), intent(in)    :: eigensolver
    integer,                intent(in)    :: unit

    type(scf_type) :: scf_info

    call push_sub("atom_solve")

    !Initialize SCF procedure
    call scf_init(atm%nspin, atm%m, scf_info)

    !Perform SCF calculations
    call scf_run(atm%m, atm%nspin, atm%wave_eq, scf_info, atm%n_states, &
                 atm%states, atm%potential, atm%xc_model, &
                 integrator_sp, integrator_dp, eigensolver, unit)

    !End SCF procedure
    call scf_end(scf_info)

    call pop_sub()
  end subroutine atom_solve

  subroutine atom_energies(atm, etot, ekin, eh, enuc, exc)
    !-----------------------------------------------------------------------!
    ! Computes the various terms of the energy.                             ! 
    !                                                                       !
    !  atm  - atom information                                              !
    !  etot - total enegy                                                   !
    !  ekin - kinetic energy                                                !
    !  eh   - electron-electron interaction energy                          !
    !  enuc - electron-nucleus interaction energy                           !
    !  exc  - exchange-correlation energy                                   !
    !-----------------------------------------------------------------------!
    type(atom_type), intent(inout) :: atm
    real(R8),        intent(out)   :: etot, ekin, eh, enuc, exc

    integer :: i
    real(R8) :: ev
    real(R8), allocatable :: density(:,:), vv(:)

    call push_sub("atom_energies")

    allocate(density(atm%m%np, atm%nspin), vv(atm%m%np))

    !Compute density
    density = state_density(atm%nspin, atm%n_states, atm%states, atm%m)

    !Compute potential energy
    select case (atm%nspin)
    case (1)
      do i = 1, atm%m%np
        vv(i) = v(atm%potential, atm%m%r(i), QN_NULL)
      end do
      ev = M_FOUR*M_PI*spline_integ(atm%m%np, atm%m%r, density(:,1)*vv*atm%m%r**2, &
                                    M_ZERO, atm%m%r(atm%m%np), 3)
    case (2)
      do i = 1, atm%m%np
        vv(i) = v(atm%potential, atm%m%r(i),  qn_type(0, 0, M_ZERO, -M_HALF, 0))
      end do
      ev = M_FOUR*M_PI*spline_integ(atm%m%np, atm%m%r, density(:,1)*vv*atm%m%r**2, &
           M_ZERO, atm%m%r(atm%m%np), 3)
      do i = 1, atm%m%np
        vv(i) = v(atm%potential, atm%m%r(i), qn_type(0, 0, M_ZERO, M_HALF, 0))
      end do
      ev = ev + M_FOUR*M_PI*spline_integ(atm%m%np, atm%m%r, density(:,2)*vv*atm%m%r**2, &
           M_ZERO, atm%m%r(atm%m%np), 3)
    end select

    !Now, the various energies
    ekin = sum(state_eigenvalue(atm%states)*state_charge(atm%states)) - ev
    enuc = M_FOUR*M_PI*spline_integ(atm%m%np, atm%m%r, -sum(density(:,:),2)*atm%z*atm%m%r, &
                                    M_ZERO, atm%m%r(atm%m%np), 3)
    call hartree_potential(atm%m, atm%nspin, density, eh = eh)
    call xc_potential(atm%xc_model, atm%m, atm%nspin, atm%n_states, atm%states, &
                      exc = exc)
    etot = ekin + enuc + eh + exc

    deallocate(density, vv)

    call pop_sub()
  end subroutine atom_energies

  function label(z)
    !-----------------------------------------------------------------------!
    ! Given an atomic number it returns the corresponding chemical symbol.  !
    !                                                                       !
    !  z     - atomic number                                                !
    !  label - chemical symbol                                              !
    !-----------------------------------------------------------------------!
    real(R8), intent(in) :: z
    character(3) :: label

    character(3), dimension(112), parameter :: table= &
         (/'H  ','He ','Li ','Be ','B  ','C  ','N  ','O  ','F  ','Ne ','Na ', &
           'Mg ','Al ','Si ','P  ','S  ','Cl ','Ar ','K  ','Ca ','Sc ','Ti ', &
           'V  ','Cr ','Mn ','Fe ','Co ','Ni ','Cu ','Zn ','Ga ','Ge ','As ', &
           'Se ','Br ','Kr ','Rb ','Sr ','Y  ','Zr ','Nb ','Mo ','Tc ','Ru ', &
           'Rh ','Pd ','Ag ','Cd ','In ','Sn ','Sb ','Te ','I  ','Xe ','Cs ', &
           'Ba ','La ','Ce ','Pr ','Nd ','Pm ','Sm ','Eu ','Gd ','Tb ','Dy ', &
           'Ho ','Er ','Tm ','Yb ','Lu ','Hf ','Ta ','W  ','Re ','Os ','Ir ', &
           'Pt ','Au ','Hg ','Tl ','Pb ','Bi ','Po ','At ','Rn ','Fr ','Ra ', &
           'Ac ','Th ','Pa ','U  ','Np ','Pu ','Am ','Cm ','Bk ','Cf ','Es ', &
           'Fm ','Md ','No ','Lr ','Rf ','Db ','Sg ','Bh ','Hs ','Mt ','Uun', &
           'Uuu','Uub'/)

    if (int(z) > 112) then
      label='unk'
    else
      label=table(int(z))
    end if

  end function label

  subroutine atom_test(ps_atm, integrator_sp, integrator_dp, eigensolver)
    !-----------------------------------------------------------------------!
    ! Test the pseudo-potentials tranferability.                            !
    !-----------------------------------------------------------------------!
    type(atom_type),        intent(inout) :: ps_atm
    type(integrator_type),  intent(inout) :: integrator_sp, integrator_dp
    type(eigensolver_type), intent(in)    :: eigensolver

    logical :: redo_scf
    integer :: tests, p, i
    integer :: wave_eq
    character(len=20) :: dir
    type(atom_type) :: ae_atm

    call push_sub("ps_atom_test")

    ASSERT(ps_atm%type == M_PS_ATOM)

    call oct_parse_logical('PPTestSCF', .false., redo_scf)
    if (redo_scf) then

      !Should we change the orbitals?
      if (oct_parse_isdef("PPTestOrbitals") /= 0) then
        do p = 1, ps_atm%nspin
          do i = 1, ps_atm%n_states
            call state_end(ps_atm%states(p, i))
          end do
        end do
        deallocate(ps_atm%states)
        ps_atm%n_states = 0

        call atom_init_states_from_block(ps_atm, "PPTestOrbitals")
      end if

      !Are we using the same wave-equation?
      if (ps_atm%wave_eq /= SCHRODINGER) then
        call oct_parse_int('WaveEquation', ps_atm%wave_eq, wave_eq)
        if (ps_atm%wave_eq == SCALAR_REL .and. wave_eq == DIRAC) then
          message(1) = "It is not possible to solve the Dirac equation for pseudopotentials"
          message(2) = "generated from a scalar relativistic calculation."
          call write_fatal(2)
        end if
        ps_atm%wave_eq = wave_eq
      end if

      !SCF cycle
      call atom_solve(ps_atm, integrator_sp, integrator_dp, eigensolver, out_tests)

      !Output wave-functions
      call atom_output(ps_atm, "tests")
    end if

    !Get the all electron atom
    call oct_parse_string("PPTestAEDir", "ae", dir)
    call atom_null(ae_atm)
    call atom_load(ae_atm, dir)

    !Now do the tests
    call oct_parse_int('PPTests', M_TEST_LD + M_TEST_DM, tests)

    if (iand(tests, M_TEST_LD).ne.0) then
      call atom_ld_test(ae_atm, ps_atm, integrator_dp)
    end if

    if (iand(tests, M_TEST_DM).ne.0) then
      call atom_dm_test(ae_atm, ps_atm)
    end if

    call atom_end(ae_atm)

  end subroutine atom_test

  subroutine atom_ld_test(ae_atm, ps_atm, integrator)
    !-----------------------------------------------------------------------!
    ! Test the pseudo-potentials tranferability by computing the            !
    ! logarithmic derivatives.                                              !
    !-----------------------------------------------------------------------!
    type(atom_type),       intent(in)    :: ae_atm, ps_atm
    type(integrator_type), intent(inout) :: integrator

    integer :: unit, p, i, iostat
    real(R8) :: r, rld, emin, emax, de, e, ae_ld, ps_ld
    character(5) :: label
    real(R8), parameter :: covalent_radius(112) = &
         !   H     He    Li    Be     B     C     N     O     F    Ne    Na
         (/ 0.60, 1.76, 2.32, 1.70, 1.55, 1.46, 1.42, 1.38, 1.36, 1.34, 2.91, &
         !   Mg    Al    Si     P     S    Cl    Ar     K    Ca    Sc    Ti   
            2.57, 2.23, 2.10, 2.00, 1.93, 1.87, 1.85, 3.84, 3.29, 2.72, 2.49, &
         !    V    Cr    Mn    Fe    Co    Ni    Cu    Zn    Ga    Ge    As
            2.31, 2.23, 2.21, 2.21, 2.19, 2.17, 2.21, 2.36, 2.38, 2.31, 2.27, &
         !   Se    Br    Kr    Rb    Sr     Y    Zr    Nb    Mo    Tc    Ru
            2.19, 2.15, 2.12, 4.08, 3.61, 3.06, 2.74, 2.53, 2.46, 2.40, 2.36, &
         !   Rh    Pd    Ag    Cd    In    Sn    Sb    Te     I    Xe    Cs
            2.36, 2.42, 2.53, 2.80, 2.72, 2.66, 2.66, 2.57, 2.51, 2.48, 4.44, &
         !   Ba    La    Ce    Pr    Nd    Pm    Sm    Eu    Gd    Tb    Dy
            3.74, 3.19, 3.12, 3.12, 3.10, 3.08, 3.06, 3.50, 3.04, 3.00, 3.00, &
         !   Ho    Er    Tm    Yb    Lu    Hf    Ta     W    Re    Os    Ir
            2.99, 2.97, 2.95, 3.29, 2.95, 2.72, 2.53, 2.46, 2.42, 2.38, 2.40, &
         !   Pt    Au    Hg    Tl    Pb    Bi    Po    At    Rn    Fr    Ra
            2.46, 2.53, 2.82, 2.80, 2.78, 2.76, 2.76, 2.74, 2.74, 0.00, 0.00, &
         !   Ac    Th    Pa     U    Np    Pu    Am    Cm    Bk    Cf    Es
            0.00, 3.12, 0.00, 2.68, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, &
         !   Fm    Md    No    Lr    Rf    Db    Sg    Bh    Hs    Mt    Uun
            0.00, 0.00, 0.00, 2.68, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, &
         !   Uuu   Uub
            0.00, 0.00 /)

    call push_sub("ps_atom_ld_test")

    ASSERT(ae_atm%type == M_AE_ATOM .and. ps_atm%type == M_PS_ATOM)

    message(1) = ""
    message(2) = "Logarithmic Derivatives:"
    call write_info(2)
    call write_info(2,unit=out_tests)

    !Read the input options
    r = covalent_radius(int(ae_atm%z))/units_in%length%factor
    call oct_parse_double("LogDerivativeRadius", r, rld)
    if (rld <= M_ZERO) then
      write(message(1), '(A,F14.6,A)') "Input: '",rld,"' is not a valid LogDerivativeRadius"
      message(2) = '(LogDerivativeRadius > 0.0)'
      call write_fatal(2)
    end if
    rld = rld*units_in%length%factor

    call oct_parse_double("LogDerivativeEnergyMin", -M_TWO, emin)
    call oct_parse_double("LogDerivativeEnergyMax",  M_TWO, emax)
    if (emin > emax) then
      message(1) =  "LogDerivativeEnergyMin must be greater than LogDerivativeEnergyMax"
      call write_fatal(1)
    end if
    emin = emin*units_in%energy%factor
    emax = emax*units_in%energy%factor

    call oct_parse_double("LogDerivativeEnergyStep", M_CENT, de)
    if (de <= M_ZERO) then
      message(1) = "LogDerivativeEnergyStep must be greater then zero"
    end if
    de = de*units_in%energy%factor

    !Output input info to the screen and to file tests/info
    write(message(1), '(2X,"Diagnostic radius: ",F6.3,1X,A)') &
         rld/units_out%length%factor, trim(units_out%length%abbrev)
    write(message(2), '(2X,"Minimum energy: ",F6.3,1X,A)') &
         emin/units_out%energy%factor, trim(units_out%energy%abbrev)
    write(message(3), '(2X,"Maximum energy: ",F6.3,1X,A)') &
         emax/units_out%energy%factor, trim(units_out%energy%abbrev)
    write(message(4), '(2X,"Energy step:    ",F6.3,1X,A)') &
         de/units_out%energy%factor, trim(units_out%energy%abbrev)
    call write_info(4)
    call write_info(4,unit=out_tests)

    call io_assign(unit)
    do p = 1, ps_atm%nspin
      do i = ps_atm%n_sc + 1, ps_atm%n_states
        ! We are not looping over the semi-core states because the logarithmic
        ! derivative should be the same for both the semi-core and valence states

        !Open file
        label = state_label(ps_atm%states(p, i))
        open(unit, file='tests/ld-'//trim(label(2:5)), status='unknown', iostat=iostat)
        if (iostat > 0) then
          write(message(1),'("Unable to open file ''tests/ld-",A,"''")') trim(label(2:5))
          call write_fatal(1)
        end if

        !Write header
        write(unit,'("# Logarithmic derivatives")')
        write(unit,'("# ")')
        write(unit,'("# Energy units: ",A)') trim(units_out%energy%name)
        write(unit,'("# Length units: ",A)') trim(units_out%length%name)
        write(unit,'("#")')
        write(unit,'("# ",53("-"))')
        write(unit,'("# |",7X,"e",7X,"|",5X,"ld_ae(e)",4X,"|",5X,"ld_pp(e)",4X,"|")')
        write(unit,'("# ",53("-"))')

        !Compute the logarithmic derivatives
        e = emin
        do
          ae_ld = state_ld(ps_atm%states(p, i), e, rld, ae_atm%wave_eq, &
                                          integrator, ae_atm%potential, ae_atm%m)
          ps_ld = state_ld(ps_atm%states(p, i), e, rld, ps_atm%wave_eq, &
                                       integrator, ps_atm%potential, ps_atm%m)
          write(unit,'(3X,ES14.7E2,3X,ES15.8E2,3X,ES15.8E2)') e/units_out%energy%factor, ae_ld, ps_ld

          if (e == emax) exit
          if (e + de > emax) then
            e = emax
          else
            e = e + de
          end if
        end do
        close(unit)

      end do
    end do

    call pop_sub()
  end subroutine atom_ld_test

  subroutine atom_dm_test(ae_atm, ps_atm)
    !-----------------------------------------------------------------------!
    ! Test the pseudo-potentials tranferability by computing the            !
    ! dipole matrix elements.                                               !
    !-----------------------------------------------------------------------!
    type(atom_type),       intent(in)    :: ae_atm, ps_atm

    integer :: p, i, f, k
    real(R8) :: ae_mif, ps_mif
    character(len=5) :: label_i, label_f
    type(state_type) :: ae_state_i, ae_state_f

    call push_sub("ps_atom_dm_test")

    ASSERT(ae_atm%type == M_AE_ATOM .and. ps_atm%type == M_PS_ATOM)

    call state_null(ae_state_i)
    call state_null(ae_state_f)

    message(1) = ""
    message(2) = "Dipole Matrix Elements:"
    write(message(3),'(6X,"States",11X,"AE",9X,"PS")')
    call write_info(3)
    call write_info(3,unit=out_tests)

    !Loop over spin
    spin_loop: do p = 1, ps_atm%nspin

      !Loop over i and f states
      i_loop: do i = 1, ps_atm%n_states - 1
        !Get ae state i
        do k = 1, ae_atm%n_states
          if (ae_atm%states(p, k) == ps_atm%states(p, i)) then
            ae_state_i = ae_atm%states(p, k)
            exit
          end if
        end do

        f_loop: do f = i + 1, ps_atm%n_states
          !Get ae state f
          do k = 1, ae_atm%n_states
            if (ae_atm%states(p, k) == ps_atm%states(p, f)) then
              ae_state_f = ae_atm%states(p, k)
              exit
            end if
          end do

          !Compute matrix elements
          ae_mif = state_dipole_matrix_element(ps_atm%m, ae_state_i, ae_state_f)
          ps_mif = state_dipole_matrix_element(ps_atm%m, ps_atm%states(p, i), &
                                               ps_atm%states(p, f))

          !Write results to screen
          label_i = state_label(ae_state_i, spin=.true.)
          label_f = state_label(ae_state_f, spin=.true.)
          select case (len_trim(label_i))
          case (2)
            write(message(1),'(5X,A," -- ",A,7X,F7.4,4X,F7.4)') &
                 trim(label_i), trim(label_f), ae_mif, ps_mif
          case (5)
            write(message(1),'(2X,A," -- ",A,4X,F7.4,4X,F7.4)') &
                 trim(label_i), trim(label_f), ae_mif, ps_mif
          end select
          call write_info(1)
          call write_info(1,unit=out_tests)

          call state_end(ae_state_f)
        end do f_loop

        call state_end(ae_state_i)
      end do i_loop

    end do spin_loop

    call pop_sub()
  end subroutine atom_dm_test

  subroutine atom_output(atm, dir)
    !-----------------------------------------------------------------------!
    ! Writes the density, the wavefunctions and the potential to the dir    !
    ! directory.                                                            !
    !-----------------------------------------------------------------------!
    type(atom_type),  intent(in) :: atm
    character(len=*), intent(in) :: dir

    integer  :: p, i

    call push_sub("atom_output") 

    if (atm%type == M_AE_ATOM .or. atm%type == M_PS_ATOM) then
      call states_output_density(atm%nspin, atm%n_states, atm%states, atm%m, dir)
      do p = 1,atm%nspin
        do i = 1,atm%n_states
          call state_output_wf(atm%states(p,i), atm%m, dir)
        end do
      end do
    end if

    call potential_output(atm%potential, dir)
    call xc_output(atm%xc_model, atm%m, atm%nspin, atm%n_states, atm%states, dir)

    call pop_sub()
  end subroutine atom_output

end module atom
