/*
 Copyright (C) 2004-2010 M. Oliveira, F. Nogueira

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 02111-1307, USA.
*/

#include <config.h>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_odeiv.h>
#include <math.h>


struct ode_type_struct
{
  void (* fdf) (double *pt, double *py, double *pf);
  double *pt, py[2], pf[2];
};


typedef struct ode_type_struct ode_type;


static void my_work_f(ode_type *p)
{
  (*(p->fdf))(p->pt, p->py, p->pf);
}


int f (double t, const double y[], double z[], void *params)
{
  ode_type *p = (ode_type *)params;
  p->pt = &t;
  p->py[0] = y[0];
  p->py[1] = y[1];
  my_work_f ( p );
  z[0] = p->pf[0];
  z[1] = p->pf[1];
  return GSL_SUCCESS;
}


int FC_FUNC_ (ode_solve, ODE_SOLVE)
     (double *t1, void **s, void **e, void **c, double *h1, double *hmin, 
      int *nstepmax, int *nstep, double x[], double y1[], double y2[], 
      void(* forfunc)(double *, double *, double *) )
{
  double t, y[2], h, sign;
  int i, status;
  ode_type p;

  p.fdf = forfunc;
  sign = (*t1-x[0])/fabs(*t1-x[0]);
  h = sign*(*h1);
  gsl_odeiv_system sys = { f, NULL, 2, &p };

  status = gsl_odeiv_evolve_reset ((gsl_odeiv_evolve *)(*e));
  if ( status != GSL_SUCCESS ) return(status);

  i = 0;
  t = x[0];
  y[0] = y1[0];
  y[1] = y2[0];
  while ( (*t1-t)*sign > 0.0 ) 
    {
      status = gsl_odeiv_evolve_apply( (gsl_odeiv_evolve *)(*e), 
	       (gsl_odeiv_control *)(*c), (gsl_odeiv_step *)(*s), &sys, &t, *t1, &h, y );
      if ( status != GSL_SUCCESS ) break;
      i += 1;
      x[i] = t;
      y1[i] = y[0];
      y2[i] = y[1];

      //Check for stepsize underflow
      if ( fabs(h) < fabs(*hmin) ) 
	{
	  status = 100;
	  break;
	}
      //Check if the integration is diverging
      if ( fabs(y[0]) > 1.0e50 || fabs(y[1]) > 1.0e50) 
	{
	  status = 101;
	  break;
	}
      //Check if the maximum number of steps has been reached
      if ( i+1 > *nstepmax )
	{
	  status = GSL_EMAXITER;
	  break;
	}

    }

  *nstep = i+1;

  return (status);
}
