!! Copyright (C) 2004-2010 M. Oliveira, F. Nogueira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!! 02111-1307, USA.

#include "global.h"

module xc
  use global
  use oct_parser
  use messages
  use splines
  use gsl_interface
  use utilities
  use units
  use io
  use output
  use mesh
  use xc_f90_types_m
  use xc_f90_lib_m
  use functionals
  use ps_io
  use states
  implicit none


                    !---Interfaces---!

  interface assignment (=)
     module procedure xc_copy
  end interface


                    !---Derived Data Types---!

  type xc_type
    integer :: nspin
    type(functional) :: functls(3)
    integer :: id_val
    integer :: corrections

    !Non-linear core-corrections
    !This is not initialized in xc_init but in xc_nlcc_init
    logical  :: nlcc
    real(R8) :: rc
    integer  :: np
    real(R8), pointer :: core_density(:)
  end type xc_type


                    !---Global Variables---!

  !Core-correcton scheme
  integer, parameter :: M_CC_NONE  = 0, &
                        M_CC_TM    = 1, &
                        M_CC_FHI   = 2

  !Corrections to the xc functionals
  integer, parameter :: XC_CORRECTION_NONE  =  0, &
                        XC_CORRECTION_REL_C =  1, &
                        XC_CORRECTION_REL_X =  2, &
                        XC_CORRECTION_SIC   =  4, &
                        XC_CORRECTION_ADSIC =  8

                    !---Public/Private Statements---!

  private
  public :: xc_type, &
            xc_null, &
            xc_init, &
            xc_nlcc_init, &
            assignment(=), &
            xc_end, &
            xc_model_output_info, &
            xc_model_save, &
            xc_model_load, &
            xc_potential, &
            xc_output, &
            xc_ps_io_set, &
            M_CC_NONE, &
            M_CC_TM, &
            M_CC_FHI, &
            XC_CORRECTION_NONE,  &
            XC_CORRECTION_REL_C, &
            XC_CORRECTION_REL_X, &
            XC_CORRECTION_SIC,   &
            XC_CORRECTION_ADSIC

contains

  subroutine xc_null(xc_model)
    !-----------------------------------------------------------------------!
    ! Nullifies and sets to zero all the components of xc_model.            !
    !-----------------------------------------------------------------------!
    type(xc_type), intent(out) :: xc_model

    integer :: i

    call push_sub("xc_null")

    xc_model%nspin = 0
    do i = 1, 3
      call functional_null(xc_model%functls(i))
    end do
    xc_model%id_val = 0
    xc_model%corrections = 0

    xc_model%nlcc   = .false.
    xc_model%rc     = M_ZERO
    xc_model%np     = 0
    nullify(xc_model%core_density)

    call pop_sub()
  end subroutine xc_null

  subroutine xc_init(nspin, xc_model)
    !-----------------------------------------------------------------------!
    ! Initializes exchange-correlation model data. The exchange-correlation !
    ! models are read from the input file.                                  !
    !-----------------------------------------------------------------------!
    integer,       intent(in)    :: nspin
    type(xc_type), intent(inout) :: xc_model

    integer :: id(2), irel

    call push_sub("xc_init")
    
    xc_model%nspin = nspin

    !Read exchange-correlation model control parameter
    !The correlation only functionals in the input file are multiplied by 1000
    call oct_parse_int('XCFunctional', XC_LDA_X + XC_LDA_C_PW*1000, xc_model%id_val)
    id(1) = xc_model%id_val/1000
    id(2) = xc_model%id_val - id(1)*1000

    !Read exchange-correlation corrections
    call oct_parse_int('XCCorrections', XC_CORRECTION_NONE, xc_model%corrections)

    !Initialize correlation
    irel = XC_NON_RELATIVISTIC
    if (iand(xc_model%corrections, XC_CORRECTION_REL_C) /= 0) irel = XC_RELATIVISTIC
    call functional_init(nspin, id(1), irel, xc_model%functls(1))

    !Initialize exchange
    irel = XC_NON_RELATIVISTIC
    if (iand(xc_model%corrections, XC_CORRECTION_REL_X) /= 0) irel = XC_RELATIVISTIC
    call functional_init(nspin, id(2), irel, xc_model%functls(2))

    !Are we doing SIC?
    if (iand(xc_model%corrections, XC_CORRECTION_SIC) /= 0) then
      call functional_init(nspin, XC_OEP_XC_SIC, irel, xc_model%functls(3))
    end if

    call pop_sub()
  end subroutine xc_init

  subroutine xc_nlcc_init(xc_model, m, rc, cc_scheme, core_density)
    !-----------------------------------------------------------------------!
    ! Initializes the non-local core-corrections part of the xc model.      !
    !                                                                       !
    !  xc_model     -  exchange-correlation model                           !
    !  m            -  mesh                                                 !
    !  rc           -  core cut-off radius                                  !
    !  cc_scheme    -  core-correction scheme to used. Valid options are    !
    !                  M_CC_TM (4th order even polynomial used in atom code !
    !                  by Jose Luis Martins) and M_CC_FHI (6th order        !
    !                  polynomial used in fhi98PP by Martin Fuchs).         !
    !  core-density -  the core density                                     !
    !-----------------------------------------------------------------------!
    type(xc_type),   intent(inout) :: xc_model
    type(mesh_type), intent(in)    :: m
    real(R8),        intent(inout) :: rc
    integer,         intent(in)    :: cc_scheme
    real(R8),        intent(in)    :: core_density(m%np, xc_model%nspin)

    integer :: i, i_rc
    logical :: monotonous
    real(R8) :: cd_rc, cdp_rc, cdpp_rc, cdppp_rc, a, b, c, d, e, f, g, h, dummy, dummy1
    real(R8), allocatable :: cd(:)
    type(spline_type) :: cd_spl

    call spline_null(cd_spl)

    xc_model%nlcc = .true.
    xc_model%rc   = rc
    xc_model%np   = m%np
    allocate(xc_model%core_density(m%np))

    !Get first, second, and third derivatives of the core density
    allocate(cd(m%np))
    cd = sum(core_density,2)
    call spline_init(cd_spl, m%np, m%r, cd, 3)
    cd_rc   = spline_eval(cd_spl, rc)
    cdp_rc  = spline_eval_deriv(cd_spl, rc)
    cdpp_rc = spline_eval_deriv2(cd_spl, rc)
    h = M_CENT
    cdppp_rc = spline_eval(cd_spl, rc + M_TWO*h) - &
               M_TWO*(spline_eval(cd_spl, rc + h) - &
               spline_eval(cd_spl, rc - h)) - &
               spline_eval(cd_spl, rc - M_TWO*h)
    cdppp_rc = cdppp_rc/(M_TWO*h**3)
    call spline_end(cd_spl)

    select case (cc_scheme)
    case (M_CC_TM)
      !Get parameters
      c = (rc*(cdpp_rc/cd_rc - (cdp_rc/cd_rc)**2) - cdp_rc/cd_rc)/(M_EIGHT*rc**3)
      b = (cdp_rc/cd_rc - M_FOUR*c*rc**3)/(M_TWO*rc)
      a = log(cd_rc) - b*rc**2 - c*rc**4

      !Compute partial core-density    
      do i = 1, m%np
        if (m%r(i) < rc) then
          xc_model%core_density(i) = exp(a + b*m%r(i)**2 + c*m%r(i)**4)
        else
          xc_model%core_density(i) = cd(i)
        end if
        if ( abs(xc_model%core_density(i)) < M_EPSILON ) xc_model%core_density(i)=M_ZERO
      end do

    case (M_CC_FHI)
      ! Insert a 4*Pi factor in the charge density and its derivatives
      cd_rc    = cd_rc*M_FOUR*M_PI
      cdp_rc   = cdp_rc*M_FOUR*M_PI
      cdpp_rc  = cdpp_rc*M_FOUR*M_PI
      cdppp_rc = cdppp_rc*M_FOUR*M_PI

      ! Find i_rc
      do i = 1, m%np
        if (m%r(i) > rc) exit
      end do
      i_rc = i

      ! Find the 5 non-zero coefficients of the polynomial
      ! Start with a = cd_rc
      a = cd_rc
      do
        ! Solve linear part
        d = M_TWENTY*(cd_rc - a)/rc**3 - M_TEN*cdp_rc/rc**2 + &
             M_TWO*cdpp_rc/rc - cdppp_rc/M_SIX
        e = (90.0_R8*(a - cd_rc) + M_FIFTY*rc*cdp_rc - M_ELEVEN*rc**2*cdpp_rc + &
             rc**3*cdppp_rc)/(M_TWO*rc**4)
        f = (72.0_R8*(cd_rc - a) - 42.0_R8*rc*cdp_rc + M_TEN*rc**2*cdpp_rc - &
             rc**3*cdppp_rc)/(M_TWO*rc**5)
        g = (M_SIXTY*(a - cd_rc) + 36.0_R8*rc*cdp_rc - M_NINE*rc**2*cdpp_rc + &
             rc**3*cdppp_rc)/(M_SIX*rc**6)

        ! Test monotonous decay
        dummy1 = a + d*m%r(i_rc)**3 + e*m%r(i_rc)**4 + &
                 f*m%r(i_rc)**5 + g*m%r(i_rc)**6
        monotonous = .true.
        do i = i_rc, 2, -1
          dummy = dummy1
          dummy1 = a + d*m%r(i-1)**3 + e*m%r(i-1)**4 + &
                   f*m%r(i-1)**5 + g*m%r(i-1)**6
          if (dummy1 < dummy) then
            monotonous = .false.
            exit
          end if
        end do
        if (monotonous) exit
        a = a*1.25_R8
      end do

      !Compute partial core-density    
      do i = 1, m%np
        if (m%r(i) < rc) then
          xc_model%core_density(i) = (a + d*m%r(i)**3 + e*m%r(i)**4 + &
                                     f*m%r(i)**5 + g*m%r(i)**6)/M_FOUR/M_PI
        else
          xc_model%core_density(i) = cd(i)
        end if
        if ( abs(xc_model%core_density(i)) < M_EPSILON ) xc_model%core_density(i) = M_ZERO
      end do

    end select

    !Output info
    write(out_pp,*)
    message(1) = ""
    message(2) = "Core Corrections:"
    select case (cc_scheme)
    case (M_CC_TM)
      message(3) = "  Scheme: Troullier-Martins scheme" 
      write(message(4),'(2x,a,f12.6)') "rc    :", rc
      write(message(5),'(2x,a)') "Polynomial coefficients:"
      write(message(6),'(4x,a,f16.10)') "a =", a
      write(message(7),'(4x,a,f16.10)') "b =", b
      write(message(8),'(4x,a,f16.10)') "c =", c
      write(message(9),'(2x,a)') "Matching at rc:"
      write(message(10),'(4x,a,f15.9,a,f15.9)') "cd(rc)   |  f0(rc)  =", &
           cd_rc,   "   | ", exp(a + b*rc**2 + c*rc**4)
      write(message(11),'(4x,a,f15.9,a,f15.9)') "cd'(rc)  |  f1(rc)  =", &
           cdp_rc,  "   | ", (M_TWO*b*rc + M_FOUR*c*rc**3)*exp(a + b*rc**2 + c*rc**4)
      write(message(12),'(4x,a,f15.9,a,f15.9)') "cd''(rc) |  f2(rc)  =", &
           cdpp_rc, "   | ", M_TWO*exp(a + b*rc**2 + c*rc**4)*(b + M_TWO*(b**2 + 3*c)*rc**2 + &
           M_EIGHT*b*c*rc**4 + M_EIGHT*c**2*rc**6)
      write(message(13),'(2x,a,f16.9)') "Integrated core charge: ", &
           spline_integ(m%np, m%r, M_FOUR*M_PI*xc_model%core_density*m%r**2, M_ZERO, m%r(m%np), 3)
      call write_info(13,20)
      call write_info(13,unit=out_pp)

    case (M_CC_FHI)
      message(3) = "  Scheme: FHI scheme" 
      write(message(4),'(2x,a,f12.6)') "rc    :", rc
      write(message(5),'(2x,a)') "Polynomial coefficients:"
      write(message(6),'(4x,a,f16.10)') "a =", a
      write(message(7),'(4x,a,f16.10)') "b =", M_ZERO
      write(message(8),'(4x,a,f16.10)') "c =", M_ZERO
      write(message(9),'(4x,a,f16.10)') "d =", d
      write(message(10),'(4x,a,f16.10)') "e =", e
      write(message(11),'(4x,a,f16.10)') "f =", f
      write(message(12),'(4x,a,f16.10)') "g =", g
      write(message(13),'(2x,a)') "Matching at rc:"
      write(message(14),'(4x,a,f15.9,a,f15.9)') "cd(rc)    |  f0(rc)  =", &
           cd_rc,    "   | ", a + d*rc**3 + e*rc**4 + f*rc**5 + g*rc**6
      write(message(15),'(4x,a,f15.9,a,f15.9)') "cd'(rc)   |  f1(rc)  =", &
           cdp_rc,   "   | ", M_THREE*d*rc**2 + M_FOUR*e*rc**3 + M_FIVE*f*rc**4 + M_SIX*g*rc**5
      write(message(16),'(4x,a,f15.9,a,f15.9)') "cd''(rc)  |  f2(rc)  =", &
           cdpp_rc,  "   | ", M_SIX*d*rc + M_TWELVE*e*rc**2 + M_TWENTY*f*rc**3 + M_THIRTY*g*rc**4
      write(message(17),'(4x,a,f15.9,a,f15.9)') "cd'''(rc) |  f3(rc)  =", &
           cdppp_rc, "   | ", M_SIX*d + 24.0_r8*e*rc + M_SIXTY*f*rc**2 + 120.0_r8*g*rc**3
      write(message(18),'(2x,a,f15.9)') "Integrated core charge: ", &
           spline_integ(m%np, m%r, M_FOUR*M_PI*xc_model%core_density*m%r**2, M_ZERO, m%r(m%np), 3)
      call write_info(18,20)
      call write_info(18,unit=out_pp)
    end select

  end subroutine xc_nlcc_init

  subroutine xc_copy(xc_out, xc_in)
    !-----------------------------------------------------------------------!
    ! Copies xc model xc_in to xc model xc_out.                             !
    !-----------------------------------------------------------------------!
    type(xc_type), intent(inout) :: xc_out
    type(xc_type), intent(in)    :: xc_in

    integer :: i

    call push_sub("xc_copy")

    call xc_end(xc_out)

    xc_out%nspin = xc_in%nspin
    do i = 1, 3
      xc_out%functls(i) = xc_in%functls(i)
    end do
    xc_out%id_val = xc_in%id_val
    xc_out%corrections = xc_in%corrections

    xc_out%nlcc  = xc_in%nlcc
    xc_out%rc    = xc_in%rc
    xc_out%np    = xc_in%np
    if (associated(xc_in%core_density)) then
      allocate(xc_out%core_density(xc_out%np))
      xc_out%core_density = xc_in%core_density
    end if

    call pop_sub()
  end subroutine xc_copy

  subroutine xc_end(xc_model)
    !-----------------------------------------------------------------------!
    ! Frees all memory associated to the xc_model.                          !
    !-----------------------------------------------------------------------!
    type(xc_type), intent(inout) :: xc_model

    integer :: i

    call push_sub("xc_end")

    xc_model%nspin = 0
    do i = 1, 3
      call functional_end(xc_model%functls(i))
    end do
    xc_model%id_val = 0
    xc_model%corrections = 0

    xc_model%nlcc = .false.
    xc_model%rc   = M_ZERO
    xc_model%np   = 0    
    if (associated(xc_model%core_density)) then
      deallocate(xc_model%core_density)
    end if

    call pop_sub()
  end subroutine xc_end

  subroutine xc_model_save(unit, xc_model)
    !-----------------------------------------------------------------------!
    ! Writes the exchange-correlation model data to a file.                 !
    !-----------------------------------------------------------------------!
    integer,       intent(in) :: unit
    type(xc_type), intent(in) :: xc_model

    integer :: i

    call push_sub("xc_model_save")

    write(unit) xc_model%nspin
    do i = 1, 3
      call functional_save(unit, xc_model%functls(i))
    end do
    write(unit) xc_model%id_val
    write(unit) xc_model%corrections

    write(unit) xc_model%nlcc
    if (xc_model%nlcc) then
      write(unit) xc_model%rc, xc_model%np
      write(unit) (xc_model%core_density(i), i=1,xc_model%np)
    end if

    call pop_sub()
  end subroutine xc_model_save

  subroutine xc_model_load(unit, xc_model)
    !-----------------------------------------------------------------------!
    ! Reads the exchange-correlation model data from a file.                !
    !-----------------------------------------------------------------------!
    integer, intent(in) :: unit
    type(xc_type), intent(inout) :: xc_model

    integer :: i

    call push_sub("xc_model_load")

    read(unit) xc_model%nspin
    do i = 1, 3
      call functional_load(unit, xc_model%functls(i))
    end do
    read(unit) xc_model%id_val
    read(unit) xc_model%corrections

    read(unit) xc_model%nlcc
    if (xc_model%nlcc) then
      read(unit) xc_model%rc, xc_model%np
      allocate(xc_model%core_density(xc_model%np))
      read(unit) (xc_model%core_density(i), i=1,xc_model%np)
    end if

    call pop_sub()
  end subroutine xc_model_load

  subroutine xc_potential(xc_model, m, nspin, n_states, states, vxc, exc, vxctau)
    !-----------------------------------------------------------------------!
    ! Given a set of states, computes the corresponding                     !
    ! exchange-correlation potentials and energies.                         !
    !                                                                       !
    !  xc_model - exchange-correlation model                                !
    !  m        - mesh                                                      !
    !  nspin    - number of spin channels                                   !
    !  n_states - the number of the states                                  !
    !  states   - the states                                                !
    !  vxc      - exchange-correlation potential                            !
    !  exc      - exchange-correlation energy                               !
    !  vxctau   - extra term arising from MGGA functionals                  !
    !-----------------------------------------------------------------------!
    type(xc_type),    intent(inout)  :: xc_model
    type(mesh_type),  intent(in)  :: m
    integer,          intent(in)  :: nspin, n_states
    type(state_type), intent(in)  :: states(nspin, n_states)
    real(R8),         intent(out), optional :: vxc(m%np, nspin)
    real(R8),         intent(out), optional :: exc
    real(R8),         intent(out), optional :: vxctau(m%np, nspin)

    integer  :: i
    real(R8) :: eaux, ip(nspin)
    real(R8), allocatable :: e(:), v(:,:), vaux(:,:), vtau(:,:)
    real(R8), allocatable :: rho(:,:), rho_grad(:,:), rho_lapl(:,:), tau(:,:)

    call push_sub("xc_potential")

    !Allocate potential and energy work arrays
    allocate(v(m%np, xc_model%nspin), vaux(m%np, nspin), e(m%np))
    allocate(vtau(m%np, xc_model%nspin))
    e = M_ZERO; eaux = M_ZERO; v = M_ZERO; vaux = M_ZERO; vtau = M_ZERO
    if (present(exc))    exc    = M_ZERO
    if (present(vxc))    vxc    = M_ZERO
    if (present(vxctau)) vxctau = M_ZERO

    !Get total electronic density and kinetic energy density
    allocate(rho(m%np, nspin), rho_grad(m%np, nspin), rho_lapl(m%np, nspin), tau(m%np, nspin))
    rho = state_density(nspin, n_states, states, m)
    rho_grad = state_density_grad(nspin, n_states, states, m)
    rho_lapl = state_density_lapl(nspin, n_states, states, m)
    tau = state_tau(nspin, n_states, states, m)
    if (xc_model%nlcc) then
      do i = 1, nspin
        rho(:, i) = rho(:, i) + xc_model%core_density/real(nspin,R8)
        rho_grad(:, i) = rho_grad(:, i) + spline_deriv(m%np, m%r, xc_model%core_density/real(nspin,R8), 3)
      end do
    end if
    where(rho <= 1e-30)
      rho = M_ZERO
      rho_grad = M_ZERO
      rho_lapl = M_ZERO
      tau = M_ZERO
    end where

    !Get ionization potential
    do i = 1, nspin
      if (sum(state_charge(states(i, :))) == M_ZERO) then
        ip(i) = M_ZERO
      else
        ip(i) = -maxval(state_eigenvalue(states(i, :)), mask = state_charge(states(i, :)) /= M_ZERO)
      end if
    end do

    !Get energy and potential
    do i = 1, 2
      call functional_get_vxc(xc_model%functls(i), m, rho, rho_grad, rho_lapl, &
                              tau, ip, v, e, vtau)
      if (present(exc)) then
        exc = exc + M_FOUR*M_PI*spline_integ(m%np, m%r, e*m%r**2, M_ZERO, m%r(m%np), 3)
      end if
      if (present(vxc)) then
        vxc = vxc + v
      end if
      if (present(vxctau)) then
        vxctau = vxctau + vtau
      end if
    end do

    if (iand(xc_model%corrections, XC_CORRECTION_ADSIC) /= 0) then
      call functional_adsic(xc_model%functls(1:2), m, nspin, rho, rho_grad, rho_lapl, tau, ip, n_states, states, vaux, eaux)
      if (present(vxc)) then
        vxc = vxc - vaux
      end if
      if (present(exc)) then
        exc = exc - eaux
      end if
    end if

    if (iand(xc_model%corrections, XC_CORRECTION_SIC) /= 0) then
      
    end if

    !Deallocate arrays
    deallocate(rho, rho_grad, rho_lapl, tau, e, v, vaux)

    call pop_sub()
  end subroutine xc_potential

  subroutine xc_model_output_info(xc_model, unit, verbose_limit)
    !-----------------------------------------------------------------------!
    ! Returns the name of the exchange-correlation functionals.             !
    !-----------------------------------------------------------------------!
    type(xc_type), intent(in) :: xc_model
    integer,       intent(in), optional :: unit, verbose_limit

    integer :: i, n_message
    character(len=10) :: family

    n_message = 2
    message(1) = ""
    message(2) = "Exchange-Correlation model:"

    do i = 1, 2
      select case (functional_family(xc_model%functls(i)))
        case (XC_FAMILY_LDA);  write(family,'(A)') "LDA"
        case (XC_FAMILY_GGA);  write(family,'(A)') "GGA"
        case (XC_FAMILY_MGGA); write(family,'(A)') "MGGA"
      end select

      select case (functional_kind(xc_model%functls(i)))
      case (XC_CORRELATION)
        n_message = n_message + 1
        write(message(n_message),'(2X,"Correlation: ",A," (",A,")")') &
                                   trim(functional_name(xc_model%functls(i))), trim(family)
      case (XC_EXCHANGE)
        n_message = n_message + 1
        write(message(n_message),'(2X,"Exchange:    ",A," (",A,")")') &
                                   trim(functional_name(xc_model%functls(i))), trim(family)
      case (XC_EXCHANGE_CORRELATION)
        n_message = n_message + 1
        write(message(n_message),'(2X,"Exchange-Correlation: ",A," (",A,")")') &
                                   trim(functional_name(xc_model%functls(i))), trim(family)
      end select
    end do

    if (iand(xc_model%corrections, XC_CORRECTION_REL_X) /= 0) then
      n_message = n_message + 1
      write(message(n_message),'(2X,"Relativistic corrections for Exchange")')
    end if
    if (iand(xc_model%corrections, XC_CORRECTION_ADSIC) /= 0) then
      n_message = n_message + 1
      write(message(n_message),'(2X,"Self-interaction correction: Averaged-Density SIC")')
    end if

   if (present(unit)) then
      call write_info(n_message, unit=unit)
    else
      if (present(verbose_limit)) then
        call write_info(n_message, verbose_limit)
      else
        call write_info(n_message)
      end if
    end if

  end subroutine xc_model_output_info

  subroutine xc_output(xc_model, m, nspin, n_states, states, dir)
    !-----------------------------------------------------------------------!
    ! Given a set of states, outputs the the corresponding                  !
    ! exchange-correlation potentials and energies.                         !
    !                                                                       !
    !  xc_model - exchange-correlation model                                !
    !  m        - mesh                                                      !
    !  nspin    - number of spin channels                                   !
    !  n_states - the number of the states                                  !
    !  states   - the states                                                !
    !-----------------------------------------------------------------------!
    type(xc_type),    intent(in) :: xc_model
    type(mesh_type),  intent(in) :: m
    integer,          intent(in) :: nspin, n_states
    type(state_type), intent(in) :: states(nspin, n_states)
    character(len=*), intent(in) :: dir

    integer  :: if, p, i, unit, iostat
    character(10) :: filename
    type(xc_type) :: xc_tmp
    real(R8), allocatable :: e(:), v(:,:)

    call push_sub("xc_output")

    !Create a temporary xc model
    call xc_null(xc_tmp)
    xc_tmp = xc_model
    do if = 1, 3
      call functional_end(xc_tmp%functls(if))
      call functional_init(nspin, 0, 0, xc_tmp%functls(if))
    end do

    !Allocate memory
    allocate(v(m%np, nspin), e(m%np))

    !Loop over the functionals
    do if = 1, 2
      if (functional_kind(xc_model%functls(if)) < 0) cycle

      !Get the potential
      xc_tmp%functls(1) = xc_model%functls(if)
      call xc_potential(xc_tmp, m, nspin, n_states, states, vxc=v)

      do p = 1, nspin
        !Get the filename
        select case (functional_kind(xc_model%functls(if)))
        case (XC_EXCHANGE)
          filename = "v_x"
        case (XC_CORRELATION)
          filename = "v_c"
        case (XC_EXCHANGE_CORRELATION)
          filename = "v_xc"
        end select
        if (nspin == 2) then
          if (p == 1) then
            filename = trim(filename)//"_dn"
          elseif (p == 2) then
            filename = trim(filename)//"_up"
          end if
        end if

        !Open file
        call io_assign(unit)
        open(unit, file=trim(dir)//"/"//trim(filename), status='unknown', iostat=iostat)
        if (iostat > 0) then
          write(message(1),'("Unable to open file ''",A,"''")') trim(dir)//"/"//trim(filename)
          call write_fatal(1)
        end if

        !Write header
        write(unit,'("# ")')
        write(unit,'("# Energy units: ",A)') trim(units_out%energy%name)
        write(unit,'("# Length units: ",A)') trim(units_out%length%name)
        write(unit,'("#")')
        write(unit,'("# ",36("-"))')
        write(unit,'("# |",8X,"r",7X,"|",7X,"v(r)",6X,"|")')
        write(unit,'("# ",36("-"))')

        !Ouput
        do i = 1, m%np
          write(unit,'(4(3X,ES15.8E2))') m%r(i)/units_out%length%factor, &
                                         v(i, p)/units_out%energy%factor
        end do

        close(unit)

      end do
    end do

    !Deallocate memory
    deallocate(v, e)
    call xc_end(xc_tmp)

    call pop_sub()
  end subroutine xc_output

  subroutine xc_ps_io_set(xc_model)
    !-----------------------------------------------------------------------!
    ! Pass the information about the exchange-correlation model to the      !
    ! ps_io module.                                                         !
    !-----------------------------------------------------------------------!
    type(xc_type), intent(in) :: xc_model

    call push_sub("xc_ps_io_set")

    call ps_io_set_xc(xc_model%id_val)
    
    if (xc_model%nlcc) then
      call ps_io_set_nlcc(xc_model%rc, xc_model%np, xc_model%core_density)
    end if

    call pop_sub()
  end subroutine xc_ps_io_set

end module xc
